/*      TSINIT.C - initialization functions for STAMP
 ***************************************************************************
 *
 * int loadStampControlFile(const char * pszFilename_in,
 *                          StampData *  pStamp_io)
 *
 * int loadStampTransferFile(const char * pszFilename_in,
 *                           StampData *  pStamp_io)
 *
 * int loadStampSynthesisFile(const char * pszFilename_in,
 *                            StampData *  pStamp_io)
 *
 * int loadStampDictCodeTables(const char * pszFilename_in,
 *                             StampData *  pStamp_io)
 *
 * void reportStampDictCodeTable(int         eDicType_in,
 *                               StampData * pStamp_in)
 *
 * int loadStampDictOrthoChanges(const char * pszFilename_in,
 *                               StampData *  pStamp_io)
 *
 ***************************************************************************
 * edit history is in version.h
 ***************************************************************************
 * Copyright 1988, 1998 by the Summer Institute of Linguistics, Inc.
 * All rights reserved.
 */
#if _MSC_VER >= 800
#include <io.h>
#include <time.h>
#endif

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include "stamp.h"

static void fix_first_code P((CodeTable * pTable_io));
static void set_ambig P((char  *str, StampData * pStamp_io));
static void set_rootdelim P((char  *str, StampData * pStamp_io));
static void show_codetab P((CodeTable * ctp,
			    char *      type,
			    FILE *      pOutputFP_in));

static const char szWhitespace_m[7] = " \t\r\n\f\v";

/******************************************************************************
 * field code table for STAMP control files
 *
 * External Internal  used  description
 * -------- --------  ----  -----------
 * \ca        A       123   Category
 * \ccl       B       123   Category Class
 * \mcl       C       123   Morpheme Class
 * \scl       D       123   String Class
 * \ap        E       123   Allomorph Property
 * \mp        F       123   Morpheme Property
 * \ambig     G       123   Ambiguity Marker (default %)
 * \rd        H       123   Root Delimiters (default < >)
 * \ir        I        2    Insertion Rule
 * \fl        J        2    Flag Rule
 * \cr        K        2    Copy Rule
 * \lc        L        23   Lexical Change
 * \maxprops  M       123   Maximum number of Allomorph/Morpheme Properties
 * \rscid     N         3   Regular Sound Change Markers (default *)
 * \rsc       R         3   Regular Sound Change
 * \sr        S        2    Substitution Rule
 * \test      T         3   Synthesis Test
 * \strcheck  V       123   Character list for string checks 1.5b BJY
 * \pcl       W       123   Punctuation Class   2.1b1 hab
 *
 * used: 1=STAMP, 2=Transfer, 3=Synthesis
 */
static const CodeTable asControlFields_m = { "\
\\ca\0A\0\
\\ccl\0B\0\
\\mcl\0C\0\
\\scl\0D\0\
\\ap\0E\0\
\\mp\0F\0\
\\ambig\0G\0\
\\rd\0H\0\
\\ir\0I\0\
\\fl\0J\0\
\\cr\0K\0\
\\lc\0L\0\
\\maxprops\0M\0\
\\rscid\0N\0\
\\rsc\0R\0\
\\sr\0S\0\
\\test\0T\0\
\\strcheck\0V\0\
\\pcl\0W\0\
", 19, NULL };			/* added \\pcl 2.1b1 hab */

/***************************************************************************
 * field code table for reading dictionary code change tables
 *
 *      External Internal       description
 *      -------- --------       -----------
 *       \ch       C            Change
 *       \prefix   P            prefix dictionary code section
 *       \infix    I            infix dictionary code section
 *       \root     R            root dictionary code section
 *       \suffix   S            suffix dictionary code section
 *       \Unified  U            unified dictionary code section
 */
static const CodeTable asDictCodeFields_m = { "\
\\ch\0C\0\
\\prefix\0P\0\
\\infix\0I\0\
\\root\0R\0\
\\suffix\0S\0\
\\unified\0U",
					      6, NULL };

/***************************************************************************
 *  field code table for orthography change file
 *
 *      External Internal       description
 *      -------- --------       -----------
 *       \ch       C            Change
 *       \scl      S            String class definition
 */
static const CodeTable asDictOrthoFields_m = { "\
\\ch\0C\0\
\\scl\0S",
					       2, NULL };

/*************************************************************************
 * NAME
 *    loadStampControlFile
 * DESCRIPTION
 *    Load the STAMP control file, storing the information contained
 *    therein.
 * RETURN VALUE
 *    zero if successful, nonzero if an error occurs
 */
int loadStampControlFile(pszFilename_in, pStamp_io)
const char *	pszFilename_in;
StampData *	pStamp_io;
{
char *	recp;		/* overall record pointer */
char *	rp;		/* Current string in record */
char *	endp;		/* End of current string in record */
int	code;		/* Field code for current string in record */
FILE *	infp;
unsigned long	uiTemp;
/*
 *  read the entire transfer control file into a single record
 */
infp = fopen(pszFilename_in, "r");
if (infp == NULL) 
    {
    return 1;
    }
recp = readStdFormatRecord(infp, &asControlFields_m, pStamp_io->cComment,
			   NULL);
if (pStamp_io->iDebugLevel && (pStamp_io->pLogFP != NULL))
    fprintf(pStamp_io->pLogFP, "\nRECORD IS:\n%s\n\n",
	    recp ? recp : "?unreadable? (NULL)");
if (recp == (char *)NULL)
    {
    return 0;			/* nothing in the file?  oh, well... */
    }
/*
 *  fill in global transfer data according to codes in record
 */
for ( rp = recp ; *rp != NUL ; rp = endp )
    {                           /* field code is first char on line */
    code = *rp++;
    rp += strspn(rp, szWhitespace_m);	/* skip over white space */
    endp = rp + strlen(rp) + 1;         /* get pointer to next field */
    if ((pStamp_io->iDebugLevel == 5) && (pStamp_io->pLogFP != NULL))
        fprintf(pStamp_io->pLogFP, "\ncode = %c\n", code);
    switch (code)
        {
        case 'A':               /* Category (\ca) */
            pStamp_io->pCategories = addAmpleCategory(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pLogFP);
            break;

        case 'B':               /* Category Class (\ccl) */
	    /* create the category class */
            pStamp_io->pCategClasses = addAmpleCategClass(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pCategClasses,
						      pStamp_io->pLogFP);
            break;

        case 'C':               /* Morpheme Class (\mcl) */
	    /* create the morpheme class */
            pStamp_io->pMorphClasses = addAmpleMorphClass(rp,
						      pStamp_io->pMorphClasses,
						      pStamp_io->pLogFP);
            break;

        case 'D':               /* String Class (\scl) */
				/* create the string class */
            pStamp_io->pStringClasses = addStringClass(rp,
						    pStamp_io->pStringClasses);
            break;

        case 'E':               /* Allomorph Property (\ap) */
            addAmpleProperty(rp, AMPLE_ALLO_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'F':               /* Morpheme Property (\mp) */
            addAmpleProperty(rp, AMPLE_MORPH_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'G':               /* ambiguity marker (\ambig) */
            set_ambig(rp, pStamp_io);
            break;

        case 'H':               /* root delimiters (\rd) */
            set_rootdelim(rp, pStamp_io);
            break;

        case 'V':       /* \strcheck string validation list 1.5b BJY */
	    if (pStamp_io->pszValidCharacters != NULL)
		freeMemory(pStamp_io->pszValidCharacters);
            pStamp_io->pszValidCharacters = setAmpleValidCharacters( rp );
            break;
            
        case 'M':       /* maximum number of properties (\maxprops) */
	    if (pStamp_io->sProperties.uiPropertyCount != 0)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP, "\n\
DECLARATIONS FILE: Must define maximum number before defining any properties\n");
		}
	    uiTemp = strtoul(rp, NULL, 10);
	    if (uiTemp >= 0x10000L)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
	      "\nDECLARATIONS FILE: Cannot have more than 65535 properties\n");
		}
	    else if (uiTemp < pStamp_io->sProperties.uiMaxProperty)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
     "\nDECLARATIONS FILE: Cannot decrease the number of properties from %u\n",
			    pStamp_io->sProperties.uiMaxProperty);
		}
	    else
		pStamp_io->sProperties.uiMaxProperty = (unsigned short)uiTemp;
            break;
            
        case 'W':               /* Punctuation Class (\pcl)  2.1b1 hab */
				/* create the punctuation class */
            pStamp_io->pPunctClasses = addPunctClass(rp,
						     pStamp_io->pPunctClasses);
            break;

        } /* end switch */

    } /* end for */
return 0;
}

/*************************************************************************
 * NAME
 *    loadStampTransferFile
 * DESCRIPTION
 *    Load the transfer control file, storing the information contained
 *    therein.
 * RETURN VALUE
 *    none
 */
int loadStampTransferFile(pszFilename_in, pStamp_io)
const char *	pszFilename_in;
StampData *	pStamp_io;
{
char *	recp;		/* overall record pointer */
char *	rp;		/* Current string in record */
char *	endp;		/* End of current string in record */
int	code;		/* Field code for current string in record */
int	i;		/* Temp loop var */
FILE *	infp;
unsigned long	uiTemp;

infp = fopen(pszFilename_in, "r");
if (infp == NULL)
    return 1;
/*
 *  read the entire transfer control file into a single record
 */
recp = readStdFormatRecord(infp, &asControlFields_m, pStamp_io->cComment,
			   NULL);
if (recp == (char *)NULL)
    return 0;			/* nothing in the file?  oh, well... */

pStamp_io->iCopyRulesTotalCount = 0;               /* Init total copy rules */
/*
 *  fill in global transfer data according to codes in record
 */
for ( rp = recp ; *rp != NUL ; rp = endp )
    {                           /* field code is first char on line */
    code = *rp++;
    rp += strspn(rp, szWhitespace_m);	/* skip over white space */
    endp = rp + strlen(rp) + 1;         /* get pointer to next field */
    switch (code)
        {
        case 'A':               /* Category (\ca) */
            pStamp_io->pCategories = addAmpleCategory(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pLogFP);
            break;

        case 'B':               /* Category Class (\ccl) */
	    /* create the category class */
            pStamp_io->pCategClasses = addAmpleCategClass(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pCategClasses,
						      pStamp_io->pLogFP);
            break;

        case 'C':               /* Morpheme Class (\mcl) */
	    /* create the morpheme class */
            pStamp_io->pMorphClasses = addAmpleMorphClass(rp,
						      pStamp_io->pMorphClasses,
						      pStamp_io->pLogFP);
            break;

        case 'D':               /* String Class (\scl) */
	    /* create the string class */
            pStamp_io->pStringClasses = addStringClass(rp,
						    pStamp_io->pStringClasses);
            break;

        case 'E':               /* Allomorph Property (\ap) */
	    addAmpleProperty(rp, AMPLE_ALLO_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'F':               /* Morpheme Property (\mp) */
	    addAmpleProperty(rp, AMPLE_MORPH_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'G':               /* ambiguity marker (\ambig) */
            set_ambig(rp, pStamp_io);
            break;

        case 'H':               /* root delimiters (\rd) */
            set_rootdelim(rp, pStamp_io);
            break;

        case 'I':               /* Insertion Rule (\ir) */
            addStampInsertionRule( rp, pStamp_io );
            break;

        case 'J':               /* Flag Rule (\flg) */
            addStampFlagRule( rp, pStamp_io );
            break;

        case 'K':               /* Copy Rule (\cr) */
            pStamp_io->eRuleType = COPYRULE;
            addStampSubstitutionRule(rp, pStamp_io);
            pStamp_io->iCopyRulesTotalCount++;     /* Count number of crules */
            break;

        case 'L':               /* Lexical Change (\lc) */
	    /*
	     *  parse the lexical change and, if successful, add it to the list
	     *  for transfer lexical changes
	     */
	    pStamp_io->pTransferLexChg = storeStampLexChange(
						    pStamp_io->pTransferLexChg,
						    rp,
						    pStamp_io);
            break;
            
        case 'M':       /* maximum number of properties (\maxprops) */
	    if (pStamp_io->sProperties.uiPropertyCount != 0)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP, "\n\
TRANSFER FILE: Must define maximum number before defining any properties\n");
		}
	    uiTemp = strtoul(rp, NULL, 10);
	    if (uiTemp >= 0x10000L)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
		  "\nTRANSFER FILE: Cannot have more than 65535 properties\n");
		}
	    else if (uiTemp < pStamp_io->sProperties.uiMaxProperty)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
	 "\nTRANSFER FILE: Cannot decrease the number of properties from %u\n",
			    pStamp_io->sProperties.uiMaxProperty);
		}
	    else
		pStamp_io->sProperties.uiMaxProperty = (unsigned short)uiTemp;
            break;

        case 'S':               /* Substitution Rule (\sr) */
            pStamp_io->eRuleType = SUBRULE;
            addStampSubstitutionRule(rp, pStamp_io);
            break;

        case 'V':       /* \strcheck string validation list 1.5b BJY */
	    if (pStamp_io->pszValidCharacters != NULL)
		freeMemory(pStamp_io->pszValidCharacters);
            pStamp_io->pszValidCharacters = setAmpleValidCharacters( rp );
            break;

        case 'W':               /* Punctuation Class (\pcl)  2.1b1 hab */
				/* create the punctuation class */
            pStamp_io->pPunctClasses = addPunctClass(rp,
						     pStamp_io->pPunctClasses);
            break;

        } /* end switch */

    } /* end for */

if ( pStamp_io->iCopyRulesTotalCount > 0 )
    {		/* If there were copy rules */
		/* Allocate an array to keep track of crule execution */
    pStamp_io->pabCopyRulesDone = (int *)allocMemory(
				pStamp_io->iCopyRulesTotalCount * sizeof(int));
    for ( i = 0 ; i < pStamp_io->iCopyRulesTotalCount ; i++ )
        pStamp_io->pabCopyRulesDone[i] = 0;
    }
return 0;
}

/*************************************************************************
 * NAME
 *    loadStampSynthesisFile
 * DESCRIPTION
 *    Load the synthesis control file, and fill in global information from
 *    the various fields.
 * RETURN VALUE
 *    none
 */
int loadStampSynthesisFile(pszFilename_in, pStamp_io)
const char *	pszFilename_in;
StampData *	pStamp_io;
{
char *	recp;		/* overall record pointer */
char *	rp;		/* Current string in record */
char *	endp;		/* End of current string in record */
int	code;		/* Field code for current string in record */
FILE *	infp;
unsigned long	uiTemp;

infp = fopen(pszFilename_in, "r");
if (infp == NULL)
    return 1;
/*
 *  read the entire synthesis control file into a single record
 */
recp = readStdFormatRecord(infp, &asControlFields_m, pStamp_io->cComment,
			   NULL);
if (recp == (char *)NULL)
    return 0;			/* nothing in the file?  oh, well... */
/*
 *  fill in global synthesis data according to codes in record
 */
for ( rp = recp ; *rp != NUL ; rp = endp )
    {                           /* field code is first char on line */
    code = *rp++;
    rp += strspn(rp, szWhitespace_m);	/* skip over white space */
    endp = rp + strlen(rp) + 1;         /* get pointer to next field */
    switch (code)
        {
        case 'A':               /* Category (\ca) */
            pStamp_io->pCategories = addAmpleCategory(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pLogFP);
            break;

        case 'B':               /* Category Class (\ccl) */
	    /* create the category class */
            pStamp_io->pCategClasses = addAmpleCategClass(rp,
						      pStamp_io->pCategories,
						      pStamp_io->pCategClasses,
						      pStamp_io->pLogFP);
            break;

        case 'C':               /* Morpheme Class (\mcl) */
	    /* create the morpheme class */
            pStamp_io->pMorphClasses = addAmpleMorphClass(rp,
						      pStamp_io->pMorphClasses,
						      pStamp_io->pLogFP);
            break;

        case 'D':               /* String Class (\scl) */
	    /* create the string class */
            pStamp_io->pStringClasses = addStringClass(rp,
						    pStamp_io->pStringClasses);
            break;

        case 'E':               /* Allomorph Property (\ap) */
	    addAmpleProperty(rp, AMPLE_ALLO_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'F':               /* Morpheme Property (\mp) */
	    addAmpleProperty(rp, AMPLE_MORPH_PROP,
			     &pStamp_io->sProperties,
			     pStamp_io->pLogFP);
            break;

        case 'G':               /* ambiguity marker (\ambig) */
            set_ambig(rp, pStamp_io);
            break;

        case 'H':               /* root delimiters (\rd) */
            set_rootdelim(rp, pStamp_io);
            break;

        case 'N':               /* regular sound change markers (\rscid) */
            setStampRegSoundChangeMarkers(rp, pStamp_io);
            break;

        case 'L':               /* Lexical Change (\lc) */
	    /*
	     *  parse the lexical change and, if successful, add it to the list
	     *  for synthesis lexical changes
	     */
	    pStamp_io->pSynthesisLexChg = storeStampLexChange(
						   pStamp_io->pSynthesisLexChg,
						   rp,
						   pStamp_io);
            break;
            
        case 'M':       /* maximum number of properties (\maxprops) */
	    if (pStamp_io->sProperties.uiPropertyCount != 0)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP, "\n\
SYNTHESIS FILE: Must define maximum number before defining any properties\n");
		}
	    uiTemp = strtoul(rp, NULL, 10);
	    if (uiTemp >= 0x10000L)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
		 "\nSYNTHESIS FILE: Cannot have more than 65535 properties\n");
		}
	    else if (uiTemp < pStamp_io->sProperties.uiMaxProperty)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
	"\nSYNTHESIS FILE: Cannot decrease the number of properties from %u\n",
			    pStamp_io->sProperties.uiMaxProperty);
		}
	    else
		pStamp_io->sProperties.uiMaxProperty = (unsigned short)uiTemp;
            break;

        case 'R':               /* Regular Sound Change (\rsc) */
            addStampRegSoundChange(rp, pStamp_io);
            break;

        case 'T':               /* Synthesis Test (\test) */
            addStampTest( rp, pStamp_io );
            break;

        case 'V':       /* \strcheck string validation list 1.5b BJY */
	    if (pStamp_io->pszValidCharacters != NULL)
		freeMemory(pStamp_io->pszValidCharacters);
            pStamp_io->pszValidCharacters = setAmpleValidCharacters( rp );
            break;

        case 'W':               /* Punctuation Class (\pcl)  2.1b1 hab */
				/* create the punctuation class */
            pStamp_io->pPunctClasses = addPunctClass(rp,
						     pStamp_io->pPunctClasses);
            break;

        } /* end switch */

    } /* end for */

deleteStampCurrentYYBuffer(); /* free buffer allocated by yylex */
return 0;
}

/*****************************************************************************
 * NAME
 *    fix_first_code
 * DESCRIPTION
 *    add the pszFirstCode code to the pCodeTable if it isn't already there
 * RETURN VALUE
 *    none
 */
static void fix_first_code(pTable_io)
CodeTable *	pTable_io;
{
unsigned	i;
char *		p;
unsigned	uiOldSize;
unsigned	uiNewSize;

if (pTable_io == NULL)
    return;
if (pTable_io->pszFirstCode == NULL)
    {
    pTable_io->pszFirstCode = duplicateString(pTable_io->pCodeTable);
    return;
    }
for ( p = pTable_io->pCodeTable, i = 0 ; i < pTable_io->uiCodeCount ; ++i )
    {
    if (strcmp(p, pTable_io->pszFirstCode) == 0)
	return;
    p += strlen(p) + 1;		/* move past match string */
    p += strlen(p) + 1;		/* move past replace string */
    }
uiOldSize = p - pTable_io->pCodeTable;
uiNewSize = uiOldSize + strlen(pTable_io->pszFirstCode) + 3;

if (pTable_io->pCodeTable == NULL)
    pTable_io->pCodeTable = allocMemory(uiNewSize);
else
    pTable_io->pCodeTable = reallocMemory(pTable_io->pCodeTable, uiNewSize);
strcpy(pTable_io->pCodeTable + uiOldSize, pTable_io->pszFirstCode);
strcpy(pTable_io->pCodeTable + uiOldSize + strlen(pTable_io->pszFirstCode)+1,
       "r");
pTable_io->uiCodeCount++;
}

/*************************************************************************
 * NAME
 *    loadStampDictCodeTables
 * DESCRIPTION
 *    Load and process the dictionary codes table file.
 *
 *    field code table for dictionary code table file
 *
 *      External Internal       description
 *      -------- --------       -----------
 *	\ch       C		code change
 *	\prefix   P		prefix region marker
 *	\infix    I		infix region marker
 *	\root     R		root region marker
 *	\suffix   S		suffix region marker
 *	\unified  U		unified dictionary region marker
 * RETURN VALUE
 *    none
 */
int loadStampDictCodeTables(pszFilename_in, pStamp_io)
const char *	pszFilename_in;
StampData *	pStamp_io;
{
char *		recp;		/* overall record pointer */
char *		rp;		/* Current entry in record */
char *		endp;		/* End of current entry in record */
char *		p;
char *		firstmark;
int		code;		/* Field code for current entry */
char *		pszMatch;
char *		pszReplace;
int		cQuote;
unsigned	uiTableLength = 0;
unsigned	uiOldLength;
CodeTable **	ppTable = NULL;
FILE *		infp;
static const char	szErrorHead_s[] = "\nDICTIONARY CODE TABLE: ";

pStamp_io->pDictTable   = NULL;
pStamp_io->pPrefixTable = NULL;
pStamp_io->pInfixTable  = NULL;
pStamp_io->pRootTable   = NULL;
pStamp_io->pSuffixTable = NULL;

infp = fopen(pszFilename_in, "r");
if (infp == NULL)
    return 1;
/*
 *  read the dictionary code table file into a single record
 */
recp = readStdFormatRecord(infp, &asDictCodeFields_m, pStamp_io->cComment,
			   NULL);
if (recp == NULL)
    return 0;			/* nothing in the file?  oh, well... */
/*
 *  fill in the dictionary code tables
 */
for ( rp = recp ; *rp != NUL ; rp = endp )
    {                           /* field code is first char on line */
    code = *rp++;
    rp += strspn(rp, szWhitespace_m);	/* skip over white space */
    endp = rp + strlen(rp) + 1;         /* get pointer to next field */
    switch (code)
        {
        case 'C':               /* dictionary code change (\ch) */
	    if (ppTable == NULL)
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
			    "%sWarning: field out of place: \\ch %s\n",
			    szErrorHead_s, rp);
		break;
		}
	    /*
	     *  isolate the pieces of the change
	     */
	    cQuote     = *rp++;
	    pszMatch   = rp;
	    pszReplace = NULL;
	    p = strchr(pszMatch, cQuote);
	    if (p != NULL)
		{
		*p++ = NUL;
		p    = strchr(p, cQuote);
		if (p != NULL)
		    {
		    pszReplace = p + 1;
		    p          = strchr(pszReplace, cQuote);
		    if (p != NULL)
			*p = NUL;
		    }
		}
	    if (    (pszMatch   == NULL) || (*pszMatch   == NUL) ||
		    (pszReplace == NULL) || (*pszReplace == NUL) )
		{
		if (pStamp_io->pLogFP != NULL)
		    fprintf(pStamp_io->pLogFP,
			    "%sInvalid \\ch field", szErrorHead_s);
		break;
		}
	    /*
	     *  add this change to the end of the table of changes
	     */
	    uiOldLength    = uiTableLength;
	    uiTableLength += strlen(pszMatch) + strlen(pszReplace) + 2;
	    if ((*ppTable)->pCodeTable == NULL)
		(*ppTable)->pCodeTable = allocMemory(uiTableLength);
	    else
		(*ppTable)->pCodeTable = reallocMemory((*ppTable)->pCodeTable,
						     uiTableLength);
	    strcpy((*ppTable)->pCodeTable + uiOldLength, pszMatch);
	    strcpy((*ppTable)->pCodeTable + uiOldLength + strlen(pszMatch)+1,
		   pszReplace);
	    (*ppTable)->uiCodeCount++;
            break;

        case 'P':               /* Prefix region marker (\prefix) */
	    ppTable = &(pStamp_io->pPrefixTable);
	    goto common;

        case 'I':               /* Infix region marker (\infix) */
	    ppTable = &(pStamp_io->pInfixTable);
	    goto common;

        case 'R':               /* Root region marker (\root) */
	    ppTable = &(pStamp_io->pRootTable);
	    goto common;

        case 'S':               /* Suffix region marker (\suffix) */
	    ppTable = &(pStamp_io->pSuffixTable);
	    goto common;

	case 'U':		/* unified dictionary codes */
	    ppTable = &(pStamp_io->pDictTable);
common:
	    *ppTable = (CodeTable *)allocMemory(sizeof(CodeTable));
	    (*ppTable)->pCodeTable = NULL;
	    uiTableLength = 0;
	    (*ppTable)->uiCodeCount  = 0;
	    firstmark = rp;
	    if (*firstmark == pStamp_io->sTextCtl.cFormatMark)
		{
		isolateWord(firstmark);
		(*ppTable)->pszFirstCode = duplicateString(firstmark);
		}
	    else
		(*ppTable)->pszFirstCode = NULL;
	    break;

	default:
	    break;
        } /* end switch */

    } /* end for */
if (pStamp_io->iDebugLevel && (pStamp_io->pLogFP != NULL))
    {
    show_codetab(pStamp_io->pDictTable,  "Unified", pStamp_io->pLogFP);
    show_codetab(pStamp_io->pPrefixTable,"Prefix",  pStamp_io->pLogFP);
    show_codetab(pStamp_io->pInfixTable, "Infix",   pStamp_io->pLogFP);
    show_codetab(pStamp_io->pRootTable,  "Root",    pStamp_io->pLogFP);
    show_codetab(pStamp_io->pSuffixTable,"Suffix",  pStamp_io->pLogFP);
    }
fix_first_code(pStamp_io->pDictTable);
fix_first_code(pStamp_io->pPrefixTable);
fix_first_code(pStamp_io->pInfixTable);
fix_first_code(pStamp_io->pSuffixTable);
fix_first_code(pStamp_io->pRootTable);

return 0;
}

/***************************************************************************
 * NAME
 *    reportStampDictCodeTable
 * DESCRIPTION
 *    Show the size of the designated code table.
 * RETURN VALUE
 *    none
 */
void reportStampDictCodeTable(eType_in, pStamp_in)
int		eType_in;
StampData *	pStamp_in;
{
static const char	szReportMessage_s[] =
		       "\t%d change%s loaded from %s dictionary code table.\n";
CodeTable *	pTable;
char *		pszType;

if ((eType_in != 0) && (pStamp_in != NULL))
    {
    switch (eType_in)
	{
	case PFX:
	    pTable  = pStamp_in->pPrefixTable;
	    pszType = "prefix";
	    break;
	case IFX:
	    pTable  = pStamp_in->pInfixTable;
	    pszType = "infix";
	    break;
	case SFX:
	    pTable  = pStamp_in->pSuffixTable;
	    pszType = "suffix";
	    break;
	case ROOT:
	    pTable  = pStamp_in->pRootTable;
	    pszType = "root";
	    break;
	default:
	    pTable  = pStamp_in->pDictTable;
	    pszType = "unified";
	    break;
	}
    if (pTable != NULL)
	reportMessage(TRUE, szReportMessage_s,
		      pTable->uiCodeCount,
		      (pTable->uiCodeCount == 1) ? "" : "s",
		      pszType );
    }
}

/*************************************************************************
 * NAME
 *    loadStampDictOrthoChanges
 * DESCRIPTION
 *    Load the dictionary orthography change file, storing the information
 *    contained therein.
 * RETURN VALUE
 *    none
 */
int loadStampDictOrthoChanges(pszFilename_in, pStamp_io)
const char *	pszFilename_in;
StampData *	pStamp_io;
{
char *		recp;		/* overall record pointer */
char *		rp;		/* Current string in record */
char *		endp;		/* End of current string in record */
int		code;		/* Field code for current string in record */
Change *	last;
Change *	chp;
FILE *		infp;

infp = fopen(pszFilename_in, "r");
if (infp == NULL)
    return 1;
/*
 *  read the entire synthesis control file into a single record
 */
recp = readStdFormatRecord(infp, &asDictOrthoFields_m, pStamp_io->cComment,
			   NULL);
if (recp == (char *)NULL)
    return 0;			/* nothing in the file?  oh, well... */
/*
 *  process the dictionary orthography changes
 */
last = (Change *)NULL;
for ( rp = recp ; *rp != NUL ; rp = endp )
    {                           /* field code is first char on line */
    code = *rp++;
    rp += strspn(rp, szWhitespace_m);	/* skip over white space */
    endp = rp + strlen(rp) + 1;         /* get pointer to next field */
    switch (code)
        {
        case 'C':                       /* change (\ch) */
	    chp = parseChangeString(rp, pStamp_io->pStringClasses);
            if (chp)
                {
                if (last)
                    last->pNext = chp;
                else
                    pStamp_io->pDictOrthoChanges = chp;
                last = chp;
                }
            break;

        case 'S':                       /* string class definition (\scl) */
					/* create the string class */
            pStamp_io->pStringClasses = addStringClass(rp,
						     pStamp_io->pStringClasses);
            break;

        } /* end switch */

    } /* end for */
return 0;
}

/*************************************************************************
 * NAME
 *    set_ambig
 * ARGUMENTS
 *    str - pointer to string
 * DESCRIPTION
 *    Use the first non-whitespace character in the string as the
 *    ambiguity marker assumed in input analysis files.  Also used
 *    in analysis files output during transfer only.
 * RETURN VALUE
 *    none
 */
static void set_ambig(str, pStamp_io)
char *		str;
StampData *	pStamp_io;
{
str += strspn(str, szWhitespace_m);
if (*str)
    pStamp_io->sTextCtl.cAnaAmbig = *str;
else if (pStamp_io->pLogFP != NULL)
    fprintf(pStamp_io->pLogFP,
	    "Invalid \ambig field - must define an ambiguity marker!\n");
}

/*************************************************************************
 * NAME
 *    set_rootdelim
 * ARGUMENTS
 *    str - pointer to a string
 * DESCRIPTION
 *    Use the first two non-whitespace characters in the string for the
 *    delimiters in the input analysis strings.
 * RETURN VALUE
 *    none
 */
static void set_rootdelim(str, pStamp_io)
char *		str;
StampData *	pStamp_io;
{
int begin, end;

str += strspn(str, szWhitespace_m);
if (*str)
    {
    begin = *str++;
    str += strspn(str, szWhitespace_m);
    if (*str)
        {
        end = *str;
        if (begin != end)
            {
            pStamp_io->cBeginRoot = begin;
            pStamp_io->cEndRoot   = end;
            return;
            }
        else
            {
	    if (pStamp_io->pLogFP != NULL)
		fprintf(pStamp_io->pLogFP,
			"The root delimiters must be different!\n");
            return;
            }
        }
    }
if (pStamp_io->pLogFP != NULL)
    fprintf(pStamp_io->pLogFP, "Invalid root delimiter field\n");
}

/***************************************************************************
 * NAME
 *    show_codetab
 * ARGUMENTS
 *    ctp  - pointer to code_table to display
 *    type - pointer to dictionary type string
 * DESCRIPTION
 *    Display the indicated dictionary code table.
 * RETURN VALUE
 *    none
 */
static void show_codetab( ctp, type, pOutputFP_in )
CodeTable *	ctp;
char *		type;
FILE *		pOutputFP_in;
{
char *	p;
int	k;

if (pOutputFP_in == NULL)
    return;
if (ctp == NULL)
    fprintf(pOutputFP_in, "Empty %s dictionary code table.\n", type);
else
    {
    if (ctp->pszFirstCode)
        fprintf(pOutputFP_in, "\n%s dictionary code table (first = %s)",
		type, ctp->pszFirstCode);
    else
        fprintf(pOutputFP_in, "\n%s dictionary code table", type);
    for ( p = ctp->pCodeTable, k = ctp->uiCodeCount ; k ; --k )
        {
        fprintf(pOutputFP_in, "\n        \"%s\"  >  \"", p);
        p += strlen(p) + 1;
        fprintf(pOutputFP_in, "%s\"", p );
        p += strlen(p) + 1;
        }
    }
}
